# Authors: Gavin Niendorf <gavinniendorf@gmail.com>
#
# Classes and methods for defining surfaces.
#
# License: MIT

import numpy as np

from .transforms import *
from .exceptions import NotOnSurfaceError
from .index import *

class geometry:
    """Class for the different surfaces in an optical system.

    Attributes
    ----------
    P : list of 3 floats/ints or just float/int
        Position of surface in the lab frame. If only one float or int is given
        then it is considered the position in the Z direction.
    D (optional): list of 3 floats/ints
        Rotation angles alpha, beta, gamma for a surface.
        Defaults to no rotation -> [0., 0., 0.]
    action : str
        Interaction of rays with the specified surface. Either reflection,
        refraction, or stop (do nothing).
    Diam : float/int
        Outer diameter of the specified surface.
    diam : float/int
        Inner diameter of the specified surface.
    N (optional): float/int
        Index of refraction of the material that the rays propagate INTO
        (meaning on the image side). If N is not given then it defaults
        to a value of 1 for air. A constant value can be specified such as 1.72.
    kappa (optional): float/int
        Defines the type of conic that the surface models.
        If kappa is None then the surface is planar.
        kappa < 0 -> hyperboloid
        kappa = 0 -> paraboloid
        0 < kappa < 1 -> hemelipsoid of revolution about major axis
        kappa = 1 -> hemisphere
        kappa > 1 -> hemelipsoid of revolution about minor axis
    c (optional): float/int
        Vertex curvature of the surface.
        If c is 0 then the surface is planar.
    name (optional): str
        Name of the surface, used for optimization
    R (generated): np.matrix((3,3))
        Rotation matrix for the surface from rotation angles D.

    """

    def __init__(self, params):
        self.P = params['P']
        self.D = np.array(params.get('D', [0., 0., 0.]))
        self.action = params['action']
        self.Diam = params['Diam']
        self.N = params.get('N', 1.)
        self.kappa = params.get('kappa', None)
        self.diam = params.get('diam', 0.)
        self.c = params.get('c', 0.)
        self.name = params.get('name', None)
        self.R = gen_rot(self.D)
        if params.get('glass'):
            self.glass = glass_index(params.get('glass'))
        self.check_params()

    def __getitem__(self, item):
        """ Return attribute of geometry. """
        return getattr(self, item)

    def __setitem__(self, item, value):
        """ Set attribute of geometry. """
        return setattr(self, item, value)

    def check_params(self):
        """Check that required parameters are given and update needed parameters.

        Summary
        -------
            If P is given as a float/int then it is converted to a np array
            with that float/int in the Z direction. If c != 0 (in the case
            of a conic) then kappa must be specified, and if kappa is greater
            than 0 then the value of c is redundant by boundary conditions of
            the conic equation. Lastly, if c == 0 in the case of a planar
            surface the None value of kappa needs to be set to a dummy value
            to avoid exceptions in calculating the conic equation. Note that
            this does not affect the calculation since c is 0.

        """

        if isinstance(self.P, float) or isinstance(self.P, int):
            #Allow on axis integer for P.
            self.P = np.array([0., 0., self.P])
        else:
            self.P = np.array(self.P)
        if self.c != 0:
            if self.kappa is None:
                raise Exception("Specify a kappa for this conic.")
            elif self.kappa>0:
                print("Warning: Specified c value is not used when kappa>0")
                self.c = np.sqrt(1/(self.kappa*pow(self.Diam/2.,2)))
        elif self.c == 0 and self.kappa is None:
            #Used for planes, does not affect calculations.
            self.kappa = 1.

    def get_surface(self, point):
        """ Returns the function and derivitive of a surface for a point. """
        return self.conics(point)

    def get_surface_plot(self, points):
        """ Returns the function value for an array of points. """
        return self.conics_plot(points)

    def conics(self, point):
        """Returns function value and derivitive list for conics and sphere surfaces.

        Note
        ----
        This differs from the conics_plot method because we are returning the function
        value from the conic at the point X,Y,Z rather than the Z value for a given X, Y
        pair. A surface is defined by the equation F(X, Y, Z) = 0, so if the ray
        intersects with the surface this function will return a function value of
        0.

        Parameters
        ----------
        point : np.array of length 3
            X, Y, and Z values for the point.

        Returns
        -------
        function : float
            Function value of the surface. A value of 0 corresponds to the ray intersecting
            the surface.
        derivitive : list of length 3
            Function derivitive of the surface at the point given. Used to determine which
            direction the ray needs to travel in and the step size to intersect the surface.

        """

        X,Y,Z = point
        rho = np.sqrt(pow(X,2) + pow(Y, 2))
        if rho > self.Diam/2. or rho < self.diam/2.:
            raise NotOnSurfaceError()
        #Conic equation.
        function = Z - self.c*pow(rho, 2)/(1 + pow((1-self.kappa*pow(self.c, 2)*pow(rho,2)), 0.5))
        #See Spencer, Murty section on rotational surfaces for definition of E.
        E = self.c / pow((1-self.kappa*pow(self.c, 2)*pow(rho,2)), 0.5)
        derivitive = [-X*E, -Y*E, 1.]
        return function, derivitive

    def conics_plot(self, point):
        """Returns Z values for an array of points for plotting conics.

        Parameters
        ----------
        point : 2d np.array with shape (n, 3)
            X, Y, and Z values for each point (for each row).

        Returns
        -------
        function : 1d np.array
            np.array of Z values for each X,Y pair.

        """

        X, Y = point[:,0], point[:,1]
        rho = np.sqrt(pow(X,2) + pow(Y,2))
        #Initialize Z value array
        function = np.zeros(len(point))
        nan_idx = (rho > self.Diam/2.) + (rho < self.diam/2.)
        rho = np.sqrt(pow(X[~nan_idx],2) + pow(Y[~nan_idx], 2))
        function[nan_idx] = np.nan
        function[~nan_idx] = self.c*pow(rho, 2)/(1 + pow((1-self.kappa*pow(self.c, 2)*pow(rho,2)), 0.5))
        return function


"""
Ray tracing and optical design module for Python
================================================

TracePy is a sequential ray tracing package written
in Python for designing optical systems in the geometric
optics regime. It features lens optimization from Scipy.
TracePy is currently in active development and any collaborators
would be welcome.

"""

from .ray import ray
from .geometry import geometry
from .optimize import optimize
from .geoplot import plotxz, plotyz, plot2d
from .optplot import spotdiagram, plotobject, rayaberration
from .iotables import *
from .transforms import *
from .raygroup import *
from.index import *


# Authors: Gavin Niendorf <gavinniendorf@gmail.com>
#
# Custom warnings and error classes used in TracePy.
#
# License: MIT

class NormalizationError(Exception):
    """ Custom exception for unnormalized input. """

class NotOnSurfaceError(Exception):
    """ Error for rays that do not intersect with a surface. """

class TraceError(Exception):
    """ Custom error for lens systems where no rays survive being traced. """



# Authors: Gavin Niendorf <gavinniendorf@gmail.com>
#
# Functions for transforming between reference frames.
#
# License: MIT

import numpy as np
from numpy import cos, sin

def gen_rot(ang):
    """Returns a rotation matrix from 3 rotation angles.

    Note
    ----
    np.matrix is deprecated and needs to be changed to a
    normal np.array but I remember difficulties with the
    regular np.array. Needs to be fixed eventually.

    Parameters
    ----------
    ang : np.array of length 3
        Euler angles alpha, beta, gamma in the lab frame.

    Returns
    -------
    np.matrix((3,3))
        Returns the rotation matrix.

    """

    alpha, beta, gamma = ang
    R_11 = cos(alpha)*cos(gamma)+sin(alpha)*sin(beta)*sin(gamma)
    R_12 = -cos(beta)*sin(gamma)
    R_13 = -sin(alpha)*cos(gamma)+cos(alpha)*sin(beta)*sin(gamma)
    R_21 = cos(alpha)*sin(gamma)-sin(alpha)*sin(beta)*cos(gamma)
    R_22 = cos(beta)*cos(gamma)
    R_23 = -sin(alpha)*sin(gamma)-cos(alpha)*sin(beta)*cos(gamma)
    R_31 = sin(alpha)*cos(beta)
    R_32 = sin(beta)
    R_33 = cos(alpha)*cos(beta)
    R = np.matrix([[R_11, R_12, R_13],\
                   [R_21, R_22, R_23],\
                   [R_31, R_32, R_33]])
    return R

def transform(R, surface, points, D=None):
    """Transforms points into the reference frame of surface.

    Note
    ----
    Arrays are flattened before return if they only have one row.

    Parameters
    ----------
    R : np.matrix((3,3))
        Rotation matrix for surface.
    surface : geometry object
        Surface whos reference frame to transform into.
    points : 2d np.array
        Point for each row that will be transformed.
    D (optional): 2d np.array
        Direction for each row that will be transformed.

    Returns
    -------
    tran_points : 2d np.array
        Points in the transformed reference frame.
    tran_D (optional): 2d np.array
        Directions in the transformed reference frame.

    """

    tran_points = np.array(np.dot(R, (points-surface.P).T).T)
    if D is not None:
        tran_D = np.array(np.dot(R, D.T).T)
        if len(tran_points) == 1:
            tran_points = tran_points.flatten()
            tran_D = tran_D.flatten()
        return tran_points, tran_D
    if len(tran_points) == 1:
        tran_points = tran_points.flatten()
    return tran_points

def lab_frame(R, surface, points, D=None):
    """Transforms points into the lab frame.

    Note
    ----
    Arrays are flattened before return if they only have one row.

    Parameters
    ----------
    R : np.matrix((3,3))
        Rotation matrix for surface.
    surface : geometry object
        Surface whos reference frame to transform from.
    points : 2d np.array
        Point for each row that will be transformed.
    D (optional): 2d np.array
        Direction for each row that will be transformed.

    Returns
    -------
    lab_points : 2d np.array
        Points in the lab reference frame.
    lab_D (optional): 2d np.array
        Directions in the lab reference frame.

    """

    lab_points = np.array(np.dot(R.T, points.T).T)+surface.P
    if D is not None:
        lab_D = np.array(np.dot(R.T, D.T).T)
        if len(lab_points) == 1:
            lab_points = lab_points.flatten()
            lab_D = lab_D.flatten()
        return lab_points, lab_D
    if len(lab_points) == 1:
        lab_points = lab_points.flatten()
    return lab_points



# Authors: Gavin Niendorf <gavinniendorf@gmail.com>
#
# Functions for plotting useful features of an optical system.
#
# License: MIT

import numpy as np

import matplotlib.pyplot as plt

from .geometry import geometry
from .transforms import transform
from .exceptions import TraceError

def _gen_object_points(surface, surface_idx, rays):
    """Transform intersection points into a surfaces' reference frame.

    Parameters
    ----------
    surface : geometry object
        Surface whos reference frame the points will be transformed into.
    surface_idx : int
        Integer corresponding to where the surface is in the propagation
        order. For example, 0 means the surface is the first surface rays
        are propagated through.
    rays : list of ray objects
        Rays that were propagated through the geometry list.

    Returns
    -------
    X, Y : np.array of len(rays)
        X, Y pair in the surface's reference frame.
    points_obj: 2d np.array
        X, Y pair points in 2d array for easy rms calculation.

    """

    points = np.array([rayiter.P_hist[surface_idx] for rayiter in rays if rayiter.P is not None])
    if points.size == 0:
        #No rays survived
        raise TraceError()
    #Get X,Y points in obj. reference frame.
    points_obj = transform(surface.R, surface, points)
    #Round arrays to upper bound on accuracy.
    points_obj = np.around(points_obj, 14)
    if points_obj.ndim == 2:
        X, Y = points_obj[:,0], points_obj[:,1]
    elif points_obj.ndim == 1:
        X, Y = points_obj[0], points_obj[1]
        points_obj = np.array([points_obj])
    return X, Y, points_obj

def spotdiagram(geo_params, rays, pltparams = {'color': 'red'}, optimizer=False):
    """Plots the transformed intersection points of rays on the stop surface.

    Note
    ----
    Directly plots the spot diagram rather than returning something.

    Parameters
    ----------
    geo_params : list of dictionaries
        Surfaces in order of propagation.
    rays : list of ray objects
        Rays that were propagated through the geometry list.
    pltparams : dictionary
        Plotting attributes of the spot diagram.
    optimizer : bool
        Flag for whether the optimizer is calling the function to get
        the rms of the spot diagram.

    """

    stop = geometry(geo_params[-1])
    X, Y, points_obj = _gen_object_points(stop, -1, rays)
    rms = np.std(points_obj[:,[0,1]] - points_obj[:,[0,1]].mean(axis=0))
    if optimizer:
        return rms
    plt.subplot(1,1,1, aspect='equal')
    plt.locator_params(axis='x', nbins=8)
    plt.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
    plt.plot(X, Y, '+', **pltparams)
    plt.text(0.65, 1.015, 'RMS=%.2E' % rms, transform=plt.gca().transAxes)

def plotobject(geo_params, rays, pltparams = {'color': 'blue'}):
    """Plots the initial ray locations in the initial objects' frame.

    Note
    ----
    Directly plots the object spot diagram rather than returning something.

    Parameters
    ----------
    geo_params : list of dictionaries
        Surfaces in order of propagation.
    rays : list of ray objects
        Rays that were propagated through the geometry list.
    pltparams : dictionary
        Plotting attributes of the object spot diagram.

    """

    start = geometry(geo_params[0])
    X, Y, points_obj = _gen_object_points(start, 0, rays)
    rms = np.std(points_obj[:,[0,1]] - points_obj[:,[0,1]].mean(axis=0))
    plt.subplot(1,1,1, aspect='equal')
    plt.ticklabel_format(style='sci', axis='both', scilimits=(0,0))
    plt.text(0.65, 1.015, 'RMS=%.2E' % rms, transform=plt.gca().transAxes)
    plt.plot(X, Y, '+', **pltparams)

def rayaberration(geo_params, rays, pltparams= {'color': 'red', 'linewidth': 2}):
    """Plots the ray aberration diagrams for an optical system.

    Note
    ----
    Directly plots the ray aberration curve rather than returning something.

    Parameters
    ----------
    geo_params : list of dictionaries
        Surfaces in order of propagation.
    rays : list of ray objects
        Rays that were propagated through the geometry list.
    pltparams : dictionary
        Plotting attributes of the curve.

    """

    start = geometry(geo_params[0])
    stop = geometry(geo_params[-1])
    X_start, Y_start, start_points = _gen_object_points(start, 0, rays)
    X_stop, Y_stop, _ = _gen_object_points(stop, -1, rays)
    #Chief ray information
    distances = np.sqrt(np.sum((start_points[:,[0,1]] - start_points[:,[0,1]].mean(axis=0))**2, axis=1))
    chief_ray_idx = np.argmin(distances)
    chief_ray = start_points[:,[0,1]][chief_ray_idx]
    #Rays along principal plane
    ray_subset_idxs = np.where(X_start == chief_ray[0])
    rel_stops = Y_stop[ray_subset_idxs] - Y_stop[chief_ray_idx]
    #Normalized pupil coordinates
    norm_coords = Y_start[ray_subset_idxs] - chief_ray[1]
    norm_coords[np.where(norm_coords < 0)] /= abs(np.min(norm_coords))
    norm_coords[np.where(norm_coords > 0)] /= abs(np.max(norm_coords))
    #Used to show axes when there is no aberration
    if np.all(rel_stops == 0.):
        plt.plot([0., 0.], [-.04, .04], 'k')
    else:
        plt.plot([0]*len(rel_stops), rel_stops, 'k')
    plt.plot(norm_coords, [0]*len(norm_coords), 'k')
    plt.plot(norm_coords, rel_stops, **pltparams)



import os

from scipy.optimize import curve_fit

import json
import yaml
import numpy as np

def cauchy_two_term(x,B,C):
    '''
    This is a simple two term cauchy for index of refraction as function of wavelength.
    https://en.wikipedia.org/wiki/Cauchy%27s_equation
    It is used to create a function for refractive index database entries when only tabulated data is available.
    '''
    return B + (C/(x**2))

def glass_index (glass):
    '''
    Given a glass name this function will return a function that takes wavelength in microns and returns index of refraction.
    - All values were taken from refractiveindexinfo's database of different optical glasses.

    https://refractiveindex.info/

    A constant value can also be input to this function to allow directly setting index

    Some glass names like F2 are shared by multiple manufacturers.
    Example input specifying the manufacturer: "F2 Schott"
    Example input without manufacturer: "F2"

    If manufacturer is not specified the following priority are used to choose the glass:
    1)schott
    2)ohara
    3)sumita
    4)cdgm
    5)hoya
    6)hikari
    7)vitron
    8)ami
    9)barberini
    10)corning
    11)lightpath
    12)lzos
    13)misc
    14)nsg

    This list isn't really a comprehensive list of manufacturers but rather the directory structure of the refractiveindexinfo yaml database.

    The functions that this function returns take a wavelength in microns and return the index of refractive index.
    '''
    #Load glass dictionary
    with open(os.path.join(os.path.dirname(__file__),os.path.join("glass","glass_dict.json")),"r") as f:
        glass_dict = eval(json.dumps(json.load(f))) #Solution to dict keys formatting
    glass_catalog = ["schott", "ohara", "sumita", "cdgm", "hoya", "hikari", "vitron", "ami", "barberini", "lightpath", "lzos", "misc", "nsg"]
    #Process input and return glass file:
    if isinstance(glass, float): #Allow index to still be set by constant value
        constant_index = float(glass)
        return lambda x=0.55: constant_index
    elif isinstance(glass, str): #Input is a glass name
        input_list = glass.split(" ")
        glass_name = input_list[0]
        if len(input_list) > 1:
            glass_manufacturer = input_list[1].lower()
        else:
            glass_manufacturer = None
        if glass_name in glass_dict:
            if glass_manufacturer:
                if glass_manufacturer in glass_dict[glass_name].keys():
                    glass_file = glass_dict[glass_name][glass_manufacturer]
                else:
                    raise Exception("{} is not in the glass catalog for {}.".format(glass_name,glass_manufacturer))
            else:
                glass_file = glass_dict[glass_name][glass_catalog[min([glass_catalog.index(x) for x in glass_dict[glass_name].keys()])]]
        else:
            raise Exception("{} is not in the glass catalog.".format(glass_name))

        #Process glass file and return the correct function for the glass (wavelength in microns)
        split_path = glass_file.split("\\") #Issue when not working on windows
        if len(split_path) == 4:
            glass_file_path = os.path.join(os.path.join(split_path[1],split_path[2]),split_path[3])
        else:
            glass_file_path = os.path.join(os.path.join(os.path.join(split_path[1],split_path[2]),split_path[3]),split_path[4])
        f = yaml.load(open(os.path.join(os.path.dirname(__file__),glass_file_path)), Loader=yaml.BaseLoader)
        type_index_function = f["DATA"][0]["type"]
        if type_index_function == 'formula 1':
            coefficents = f["DATA"][0]['coefficients'].split(" ")
            coefficents = [float(x) for x in coefficents]
            return lambda x=0.55: (1+coefficents[0]+coefficents[1]/(1-(coefficents[2]/x)**2)+coefficents[3]/(1-(coefficents[4]/x)**2))**.5
        if type_index_function == 'formula 2':
            coefficents = f["DATA"][0]['coefficients'].split(" ")
            coefficents = [float(x) for x in coefficents]
            if len(coefficents) >=7:
                return lambda x=0.55: (1+coefficents[1]/(1-coefficents[2]/x**2)+coefficents[3]/(1-coefficents[4]/x**2)+coefficents[5]/(1-coefficents[6]/x**2))**.5
            else:
                return lambda x=0.55: (1+coefficents[1]/(1-coefficents[2]/x**2)+coefficents[3]/(1-coefficents[4]/x**2))**.5
        if type_index_function == 'formula 3':
            coefficents = f["DATA"][0]['coefficients'].split(" ")
            coefficents = [float(x) for x in coefficents]
            return lambda x=0.55: (coefficents[0]-coefficents[1]*x**coefficents[2]+coefficents[3]*x**coefficents[4]+coefficents[5]*x**coefficents[6]+coefficents[7]*x**coefficents[8]+coefficents[9]*x**coefficents[10])**.5
        if type_index_function == 'formula 5':
            coefficents = f["DATA"][0]['coefficients'].split(" ")
            coefficents = [float(x) for x in coefficents]
            return lambda x=0.55: coefficents[0]-coefficents[1]*x**coefficents[2]+coefficents[3]*x**-coefficents[4]
        if type_index_function == 'tabulated n':
            raw_data = f["DATA"][0]['data'].split('\n')[:-1]
            wavelength = np.array([float(x.split(' ')[0]) for x in raw_data])
            index = np.array([float(x.split(' ')[1]) for x in raw_data])
            popt, _ = curve_fit(cauchy_two_term, wavelength, index) #Curve fit cauchy two term function
            return lambda x=0.55: popt[0] + (popt[1]/(x**2))
        if type_index_function == 'tabulated nk':
            raw_data = f["DATA"][0]['data'].split('\n')[:-1]
            wavelength = np.array([float(x.split(' ')[0]) for x in raw_data])
            index = np.array([float(x.split(' ')[1]) for x in raw_data])
            popt, _ = curve_fit(cauchy_two_term, wavelength, index) #Curve fit cauchy two term function
            return lambda x=0.55: popt[0] + (popt[1]/(x**2)) #Many of these materials are IR glass so the default0.55 is way out of the tabulated range...


# Authors: Gavin Niendorf <gavinniendorf@gmail.com>
#
# Classes and methods for defining rays and their propagation rules.
#
# License: MIT

import numpy as np

from .transforms import *
from .exceptions import NormalizationError, NotOnSurfaceError

class ray:
    """Class for rays and their propagation through surfaces.

    Note
    ----
    Also checks whether the direction cosines are normalized.

    Attributes
    ----------
    P : np.array of 3 floats/ints
        Position of ray in the lab frame.
    D : np.array of 3 floats/ints
        Direction cosines for the ray in the lab frame.
    P_hist : list of P np.arrays
        Previous P np.arrays in a list.
    D_hist : list of D np.arrays
        Previous D np.arrays in a list.
    N : float/int
        Index of refraction of current material.
    wvl: float/int
        Wavelength of the ray in microns 550nm --> 0.55.
    """

    def __init__(self, params, N_0=1):
        self.P = np.array(params['P'])
        self.D = np.array(params['D'])
        self.P_hist = [self.P]
        self.D_hist = [self.D]
        self.N = N_0
        self.wvl = params.get('wvl',0.55) #Added default wavelength 550nm
        if abs(np.linalg.norm(self.D)-1.) > .01:
            #Ray direction cosines are not normalized.
            raise NormalizationError()

    def transform(self, surface):
        """ Updates position and direction of a ray to obj coordinate system. """
        self.P, self.D = transform(surface.R, surface, np.array([self.P]), np.array([self.D]))

    def find_intersection(self, surface):
        """Finds the intersection point of a ray with a surface.

        Note
        ----
        Directly changes the self.P (position) attribute of the ray
        that corresponds to the intersection point. Also be aware
        that my error definition is different from Spencer's paper.
        I found that the more direct error equation of abs(F) allows
        me to tune my max error values to get better accuracy.

        Parameters
        ----------
        surface : geometry object
            Surface to find intersection of ray with.

        """

        #Initial guesses, see Spencer, Murty for explanation.
        s_0 = -self.P[2]/self.D[2]
        X_1 = self.P[0]+self.D[0]*s_0
        Y_1 = self.P[1]+self.D[1]*s_0
        s_j = [0., 0.]
        #Initial error.
        error = 1.
        n_iter = 0
        #Max iterations allowed.
        n_max = 1e4
        while error > 1e-6 and n_iter < n_max:
            X, Y, Z = [X_1, Y_1, 0.]+np.dot(self.D, s_j[0])
            try:
                #'normal' is the surface direction numbers.
                func, normal= surface.get_surface([X, Y, Z])
                deriv = np.dot(normal, self.D)
                #Newton-raphson method
                s_j = s_j[1], s_j[1]-func/deriv
            except NotOnSurfaceError:
                self.P = None
                return None
            #Error is how far f(X, Y, Z) is from 0.
            error = abs(func)
            n_iter += 1
        if n_iter == n_max or s_0+s_j[0] < 0 or np.dot(([X, Y, Z]-self.P), self.D) < 0.:
            self.P = None
        else:
            self.normal = normal
            self.P = np.array([X, Y, Z])

    def interact(self, surface, typeof):
        """Updates new direction of a ray for a given interaction type.

        Note
        ----
        High level method that calls the appropriate method for a given
        interaction.

        Parameters
        ----------
        surface : geometry object
            Surface to find intersection of ray with.
        typeof : str
            Type of interaction
            reflection -> Reflect the ray off the surface.
            refraction -> Refract the ray into the surface.
            stop -> Don't change ray direction.

        """
        if hasattr(surface,'glass'):
            mu = self.N / surface.glass(self.wvl)
        else:
            mu = self.N / surface.N
        a = mu*np.dot(self.D, self.normal)/pow(np.linalg.norm(self.normal), 2)
        b = (pow(mu,2)-1)/pow(np.linalg.norm(self.normal), 2)
        if typeof == 'stop':
            pass
        #Needed for total internal reflection even if typeof is refraction.
        elif b > pow(a, 2) or typeof == 'reflection':
            self.reflection(surface, a/mu)
        elif typeof == 'refraction':
            self.refraction(surface, mu, a, b)

    def reflection(self, surface, a):
        """Reflects the ray off a surface and updates the ray's direction.

        Note
        ----
        This method computes D exactly rather than numerically like in the
        refraction method.

        Parameters
        ----------
        surface : geometry object
            Surface to reflect from.
        a : float/int
            Constant defined in the interact method.

        """

        k, l, m = self.D
        K, L, M = self.normal
        self.D = np.array([k-2.*a*K, l-2.*a*L, m-2.*a*M])

    def refraction(self, surface, mu, a, b):
        """Simulates refraction of a ray into a surface and updates the ray's direction.

        Note
        ----
        My error definition is not in Spencer and Murty's paper but is inspired by my
        unique intersection error definition. We are solving for roots of a quadratic and
        I am defining my error by how far the quadtratic is from 0. See Spencer, Murty for
        derivation of the quadratic.

        Parameters
        ----------
        surface : geometry object
            Surface to refract into.
        mu, a, b : float/int
            Constants defined in the interact method.

        Returns
        -------
        0
            Returns 0 if the number of iterations exceeds the max allowed to converge.

        """

        k, l, m = self.D
        K, L, M = self.normal
        G = [-b/(2*a), -b/(2*a)]
        #Initial error.
        error = 1.
        niter = 0
        #Max iterations allowed.
        nmax = 1e5
        while error > 1e-15 and niter < nmax:
            #Newton-raphson method
            G = G[1], (pow(G[1],2)-b)/(2*(G[1]+a))
            #See Spencer, Murty for where this is inspired by.
            error = abs(pow(G[1],2)+2*a*G[1]+b)
            niter += 1
        if niter==nmax:
            self.P = None
            return 0.
        #Update direction and index of refraction of the current material.
        self.D = np.array([mu*k+G[1]*K,mu*l+G[1]*L,mu*m+G[1]*M])
        if hasattr(surface,'glass'):
            self.N = surface.glass(self.wvl)
        else:
            self.N = surface.N

    def ray_lab_frame(self, surface):
        """ Updates position and direction of a ray in the lab frame. """
        self.P, self.D = lab_frame(surface.R, surface, np.array([self.P]), np.array([self.D]))

    def update(self):
        """ Updates the P_hist and D_hist arrays from current P and D arrays. """
        self.P_hist.append(self.P)
        self.D_hist.append(self.D)

    def propagate(self, surfaces):
        """Propagates a ray through a given surfaces list.

        Note
        ----
        If self.P is None then the ray failed to converge or
        took too many iterations to meet the required accuracy.
        Note that this is used (self.P is None) as a flag in
        many other functions in TracePy.

        Parameters
        ----------
        surfaces : list of geometry objects
            Surfaces to propagate through in order of propagation.

        """

        for surface in surfaces:
            self.transform(surface)
            self.find_intersection(surface)
            #Results from failure to converge.
            if self.P is None:
                break
            self.interact(surface, surface.action)
            #Results from too many iterations.
            if self.P is None:
                break
            self.ray_lab_frame(surface)
            #Update current to history arrays.
            self.update()



# Authors: Gavin Niendorf <gavinniendorf@gmail.com>
#
# Functions for defining groups of rays.
#
# License: MIT

import numpy as np
from numpy import pi, sqrt

from .ray import ray
from .geometry import geometry

def ray_plane(geo_params, pos, radius, d, nrays=100):
    """Creates a plane of rays and propagates them through geometry.

    Note
    ----
    This is done by creating a square grid with side length 2 x radius
    and then chopping off all points outside of a radius. But this means
    the number of rays is significantly less than the parameter nrays in
    some cases. This needs to be fixed eventually.

    Parameters
    ----------
    geo_params : list of dictionaries
        Dictionaries correspond to surfaces.
    pos : list of length 3 or float/int
        X, Y, and Z values of the plane or just the Z value.
    d : list of length 3
        Direction cosines for each ray.
    radius : float/int
        Radius of ray plane.
    nrays : int
        Number of rays in the ray plane.

    Returns
    -------
    list of ray objects
        Returns rays after propagating them through geometry list.

    """

    x_mesh = np.linspace(-radius, radius, int(4./pi*sqrt(nrays)))
    y_mesh = np.linspace(-radius, radius, int(4./pi*sqrt(nrays)))
    x_points, y_points = np.meshgrid(x_mesh, y_mesh)
    xs, ys = x_points.ravel(), y_points.ravel()
    dis = sqrt(pow(xs,2) + pow(ys, 2))
    if isinstance(pos, list):
        x_circ, y_circ = xs[dis < radius] + pos[0], ys[dis < radius] + pos[1]
        #To-do: Should transform for d
        z_circ = [pos[2]]*len(x_circ)
    else:
        x_circ, y_circ = xs[dis < radius], ys[dis < radius]
        #To-do: Should transform for d
        z_circ = [pos]*len(x_circ)
    P_arr = np.vstack((x_circ, y_circ, z_circ)).T
    D_arr = np.array([d] *len(x_circ))
    #Initialize rays.
    rays = [ray(params={'D':D, 'P':P}) for D,P in zip(D_arr, P_arr)]
    geo = [geometry(surf) for surf in geo_params]
    #Propagate rays through geometry.
    for rayiter in rays:
        rayiter.propagate(geo)
    return rays
